/*
** x10d.c (c) Copyright Daniel D. Lanciani 1995-1997
** All rights reserved.
** 
** x10d.c is licensed free of charge for personal and internal
** business use only.  x10d.c may not be distributed for profit,
** nor may it be included in products or otherwise distributed by
** commercial entities to their clients or customers without the
** prior written permission of the author.
** 
** These copyright, license, and disclaimer notices must be included
** with all copies of x10d.c.
*/

/*
**
** $Id: x10d.c,v 1.1 1997/12/20 06:35:56 ncherry Exp $
**
** $Log: x10d.c,v $
** Revision 1.1  1997/12/20 06:35:56  ncherry
** Initial revision
**
**
*/


/*
** RCS Id macro
*/

#ifdef lint
# define RCS_Id(i)
#else  						/* lint */
# ifdef __GNUC__
#  define ATTRIB_UNUSED __attribute__ ((unused))
# else 						/* __GNUC__ */
#  define ATTRIB_UNUSED
# endif 					/* __GNUC__ */
# define RCS_Id(i) static char const *RCS_Id ATTRIB_UNUSED = (i);
#endif 						/* lint */

RCS_Id("$Id$")

#include <stdio.h>		/*  */
#include <errno.h>		/* from original */
#include <netdb.h>		/*  */
#include <signal.h>		/*  */
#include <syslog.h>		/*  */
#include <unistd.h>

#include <sys/types.h>		/*  */
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/time.h>		/*  */
#include <sys/ioctl.h>		/*  */
#include <sys/socket.h>		/*  */

#include <netinet/in.h>		/*  */
#include <arpa/inet.h>		/*  */

#include <termios.h>

struct termios oldsb, newsb;

extern SignalHandler fini();

extern void help(int);

int    lsocket, cm11, nclient, maxclient = 20, on = 1, off = 0;
char   rbuf[128], buf[128];

struct linger linger = { 1, 0 };

static struct sockaddr_in sin = { AF_INET };
static fd_set xprobe;

#define FD_NULL		(fd_set *) 0
#define TIMEVAL_NULL	(struct timeval *) 0

char *ctime(), *memchr(), *malloc(), *realloc();

struct clients {
	int s, cnt, flags;
	char buf[128];
} *clients;

#define C_QUIET 0x0001

#define X10_ALL_UNITS_OFF	17
#define X10_ALL_LIGHTS_ON	18
#define X10_ON			19
#define X10_OFF			20
#define X10_DIM			21
#define X10_BRIGHT		22
#define X10_EXTENDED		24

/*
** Interface Poll Signal         0x5A
** PC response to the Poll       0xC3
** Fast Macro Download           0xA5 ?
** PC Response to Macro D/L Poll 0xFB
** Enable RI                     0xEB
** Disable RI                    0xDB
** Checksum Correct              0x00
** Interface Ready               0x55
** Time Request                  0xA5
** Status request                0x8B
**
*/

#define ACK		"\0"	/* 0x00 Checksum ACK              */
#define READY           0x55    /* 0x55 Interface Ready           */

#define POLL            0x5A    /* 0x5A Interface Poll Signal     */
#define POLL_ACK        "\xC3"  /* 0xC3 PC response to the Poll   */
#define MACRO_RESP      0xFB    /* PC response to Macro poll */
#define FastMacro       0xA5    /* Fast Macro Download       */
#define PFail           0xA5    /* Power Fail                */
#define STAT_REQST      0x8B    /* Status Request            */
#define EEPROM_EVENT	0x5B	/* EEPROM address ???        */
#define TIME_DL		0x9B	/* Response to a POLL        */

#define EnableRI        0xEB    /* Enable Ring Indicator     */
#define DisablRI        0xDB    /* Disable Ring Indicator    */

#define FFail		0xF3	/* Filter fail               */
#define CPPFail		0xA6    /* CP10 Power Fail           */

struct x10aliases {
  char *name;
  int code;
} x10aliases[] = {
  { "ALLUNITSOFF", X10_ALL_UNITS_OFF },
  { "ALLLIGHTSON", X10_ALL_LIGHTS_ON },
  { "ON",          X10_ON },
  { "OFF",         X10_OFF },
  { "DIM",         X10_DIM },
  { "BRIGHT",      X10_BRIGHT },
  { 0, 0 }
};

char *x10longnames[] = {
  "",
  "Unit 1",
  "Unit 2",
  "Unit 3",
  "Unit 4",
  "Unit 5",
  "Unit 6",
  "Unit 7",
  "Unit 8",
  "Unit 9",
  "Unit 10",
  "Unit 11",
  "Unit 12",
  "Unit 13",
  "Unit 14",
  "Unit 15",
  "Unit 16",
  "All Units Off",
  "All Lights On",
  "On",
  "Off",
  "Dim",
  "Bright",
  "All Lights Off",
  "Extended Code",
  "Hail Request",
  "Hail Acknowledge",
  "Preset Dim 0",
  "Preset Dim 1",
  "Extended Data",
  "Status is On",
  "Status is Off",
  "Status Request"
};

struct cm11_setclock {
  char cmd;	/* 0x9b           */
  char seconds;
  char minutes;	/* 0-119          */
  char hours;	/* 0-11 (hours/2) */
  char yearday;	/* really 9 bits  */
  char daymask;	/* really 7 bits  */
  char house;	/* 0:timer purge, 1:monitor clear, 3:battery clear */
};

#ifndef ServiceName
#define ServiceName	"x10d"
#endif

#ifndef X10_DEV
#define X10_DEV		"/dev/cm11"
#endif

#define X10_TTY		"/dev/tty"

#ifndef Access
#define Access		"/usr/local/lib/x10d/x10d.access"
#endif

#ifndef X10Log
#define X10Log		"/usr/local/lib/x10d/x10d.log"
#endif

#define MASK(s)		(0x000000FF & s)

int ptty;

main(argc, argv)
char **argv;
{
  register FILE *m;		/*  */
  register char *p;		/*  */
  register int i, c;		/*  */

  int l;			/*  */

  struct servent *sp;		/*  */
  struct in_addr in;

  fd_set rset, probe;		/*  */

  if(fork())
    exit(0);

  /*
  ** Close the fd's so we are not associate with any controlling tty
  */

  close(0);
  close(1);
  close(2);

  /*
  ** What is this?
  */

  if(open("/", 0)) {
    syslog(LOG_CRIT, "x10d: open / not 0: %m");
    exit(1);
  }

  dup2(0, 1);
  dup2(0, 2);

  signal(SIGPIPE, SIG_IGN);

  openlog("x10d", LOG_PID, LOG_LOCAL0);

  syslog(LOG_LOCAL0, "X10D starting");

  if(!(clients = (struct clients *)
       malloc(maxclient * sizeof(struct clients)))) {
    syslog(LOG_CRIT, "no memory for clients");
    exit(1);
  }

  FD_ZERO(&rset);
  FD_ZERO(&probe);

  if(!(sp = getservbyname(ServiceName, "tcp"))) {
    syslog(LOG_ERR, "x10d: unknown service");
    exit(1);
  }

  if((lsocket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    syslog(LOG_ERR, "socket: %m");
    exit(1);
  }

  sin.sin_port = sp->s_port;

  if(bind(lsocket,(void *) &sin, sizeof(sin))) {
    syslog(LOG_ERR, "bind: %m");
    exit(1);
  }

  listen(lsocket, 5);
  ioctl(lsocket, FIONBIO, &on);
  FD_SET(lsocket, &rset);

  if((c = open(X10_TTY, O_RDWR)) >= 0) {
    ioctl(c, TIOCNOTTY, 0);
    close(c);
  }

  /* *********************************************************************
  **
  ** Open the CM11 device and set it up
  */

  if((cm11 = open(X10_DEV, O_RDWR)) < 0) {
    syslog(LOG_ERR, "%s: %m", X10_DEV);
    exit(1);
  }

  /*
  ** This works for cm11 (NJC)
  */

  {
    int s;
    s = tcgetattr(cm11, &oldsb);

    if (s < 0) {
      perror("ttopen tcgetattr");
      exit(1);
    }

    newsb = oldsb;

    newsb.c_iflag = IGNBRK | IGNPAR;
    newsb.c_oflag = 0;
    newsb.c_lflag = ISIG;
    newsb.c_cflag = (CLOCAL | B4800 | CS8 | CREAD);
    for (s = 0; s < NCC; s++)
      newsb.c_cc[s] = 0;
    newsb.c_cc[VMIN]   = 1;
    newsb.c_cc[VTIME]  = 0;

    tcsetattr(cm11, TCSADRAIN, &newsb);
  }

  FD_SET(cm11, &rset);

  /***********************************************************************/

  while(1) {
    probe = rset;

    if(select(FD_SETSIZE, &probe, FD_NULL, FD_NULL, TIMEVAL_NULL) <= 0)
    {
      syslog(LOG_ERR, "select: %m");
      sleep(2);
      continue;
    }

    if(FD_ISSET(lsocket, &probe)) {

      l = sizeof(sin);

      if((c = accept(lsocket,(struct sockaddr *) &sin, &l)) < 0)
      {
	syslog(LOG_ERR,"Bad Accept");
      }
      else
      {
	ioctl(c, FIONBIO, &on);

	if(m = fopen(Access, "r")) {
	  strcpy(buf,  "");	/* because these are used elsewhere clear them */
	  strcpy(rbuf, "");

	  while(fscanf(m, "%s %s\n", buf, rbuf) == 2)
	  {
	    if((sin.sin_addr.s_addr & inet_addr(rbuf)) == inet_addr(buf))
	    {
	      break;
	    }
	  }
	  fclose(m);

	  if((sin.sin_addr.s_addr & inet_addr(rbuf)) != inet_addr(buf))
	  {
	    close(c);
	    in.s_addr = (sin.sin_addr.s_addr & inet_addr(rbuf));
	    syslog(LOG_ERR,"Bad Accept, SIN & RBUF (%s) != BUF (%s)", inet_ntoa(in), buf);
	    goto badaccept;
	  }
	} /* end if(m = fopen(Access, "r")) */

	if(nclient >= maxclient) {
	  maxclient += 20;
	  if(!(clients = (struct clients *)realloc((char *)clients, maxclient *
						   sizeof(struct clients)))) {
	    syslog(LOG_CRIT,"out of client memory");
	    exit(1);
	  }
	}

	setsockopt(c, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger));
	clients[nclient].flags = clients[nclient].cnt = 0;
	clients[nclient++].s = c;
	FD_SET(c, &rset);
	in.s_addr = sin.sin_addr.s_addr;
	syslog(LOG_DEBUG,"Connect from %s", inet_ntoa(in));
      } /* end if(accept() < 0) */

    badaccept:;

    } /* end if(FD_ISSET(locket, &probe)) */

    for(i = 0; i < nclient; i++)
    {
      if(FD_ISSET(clients[i].s, &probe)) {
	c = sizeof(clients[i].buf) - clients[i].cnt - 1; /* Calc buf size */
	if(c <= 0)
	  goto badclient;	/* no buffer space */

	c = read(clients[i].s,
		 clients[i].buf + clients[i].cnt, c);

	if(c < 0)
	{
	  if(errno != EWOULDBLOCK)
	    goto badclient;
	  continue;		/* nothing to read */
	}

	if(c == 0) {		/* buffer closed */

	badclient:

	  syslog(LOG_DEBUG,"Client closed");
	  close(clients[i].s);
	  FD_CLR(clients[i].s, &rset);
	  nclient--;
	  for(c = i; c < nclient; c++)
	    clients[c] = clients[c + 1];
	  i--;
	  continue;
	}

	clients[i].cnt += c;
	while(p = memchr(clients[i].buf, '\n', clients[i].cnt)) {
	  c = p - clients[i].buf + 1;
	  bcopy(clients[i].buf, rbuf, c);
	  if(clients[i].cnt -= c)
	    bcopy(clients[i].buf + c, clients[i].buf, clients[i].cnt);

	  for(p = buf, l = c, c = 0; c < l; c++) {

	    if(rbuf[c] < ' ') {
	      *p = 0;
	      break;
	    }

	    if(rbuf[c] == ' ')
	      continue;

	    if(rbuf[c] >= 'a' && rbuf[c] <= 'z')
	      *p++ = rbuf[c] - 'a' + 'A';
	    else
	      *p++ = rbuf[c];
	  }/* end of for(p = buf, l = c, c = 0; c < l; c++) */

	  parse(buf, i);

	} /* End while() */
      } /* End if(FD_ISSET()) */
    } /* End of for(i = 0; i < nclient; i++) */

    if(FD_ISSET(cm11, &probe)) {

      c = x10read(cm11, rbuf, sizeof(rbuf) - 2);

      if(c < 0) {
	if(errno != EWOULDBLOCK) {
	  syslog(LOG_ERR, "read x10: %m");
	  sleep(1);
	}
	/* goto badread; */
      }
      else if(c == 0) {
	syslog(LOG_ERR, "read x10: EOF");
	sleep(1);
	/* goto badread; */
      }
      else
      {
	for(i = 0; i < c; i++)
	{
	  if((rbuf[i] & 0xff) == EEPROM_EVENT) {
	    addrreceive(rbuf + i + 1, c - i - 1);
	    if((c -= i + 3) <= 0)
	      goto badread;
	    bcopy(rbuf + i + 3, rbuf, c);
	    i = -1;
	  }
	}

	switch(rbuf[c - 1] & 0xff) {

	  case POLL:		/* receive poll */
	    pollreceive();
	    break;

#ifdef CP10
	  case FFail:		/* filter fail */
	    x10write(cm11, FFail, 1); /* 1111 0011 F3*/
	    syslog(LOG_WARNING, "filter failure");
	    break;
#endif
	  case PFail:		/* power fail */
	    setcm11clock();
	    syslog(LOG_NOTICE, "power failed");
	    break;

	  case EEPROM_EVENT:	/* eeprom address */
#ifdef CP10
	  case CPPFail:		/* CP10 power fail */
#endif
	  default:
	    syslog(LOG_ERR, "unexpected poll: %x", *rbuf & 0xff);
	  break;
	}
      }
      badread: ;
    } /* End if(FD_ISSET(cm11...) */
  } /* End while(1) */
}

static char fmap[] = { 13,  5,   3, 11,
		       15,  7,   1,  9,
		       14,  6,   4, 12,
		       16,  8,   2, 10 };

static char tmap[] = { 12,  6,  14,  2,
		       10,  1,   9,  5,
		       13,  7,  15,  3,
		       11,  0,   8,  4, 12};

addrreceive(buf, cnt)
register int cnt;
char *buf;
{
  register int c;
  static struct timeval tv;

  while(cnt < 2) {
    FD_SET(cm11, &xprobe);
    tv.tv_sec = 2;
    c = select(cm11 + 1, &xprobe, FD_NULL, FD_NULL, &tv);

    if(c < 0) {
      syslog(LOG_ERR, "select addrreceive: %m");
      return;
    }

    if(c == 0) {
      syslog(LOG_ERR, "select addrreceive timeout");
      return;
    }

    c = x10read(cm11, buf + cnt, 2 - cnt);

    if(c <= 0) {
      if(c) {
	if(errno == EWOULDBLOCK)
	  continue;
	syslog(LOG_ERR, "read addrreceive: %m");
      }
      else
	syslog(LOG_ERR, "read addrreceive: EOF");
      return;
    }
    cnt += c;
  }
}

pollreceive()
{
  register int c, i, j, r;
  int ecnt;
  static struct timeval tv;

top:
  ecnt = 0;
  x10write(cm11, POLL_ACK, 1);

again:

  FD_SET(cm11, &xprobe);
  tv.tv_sec = 2;

  c = select(cm11 + 1, &xprobe, FD_NULL, FD_NULL, &tv);

  if(c < 0) {
    syslog(LOG_ERR, "select pollreceive: %m");
    return;
  }
  if(c == 0) {
    syslog(LOG_ERR, "select pollreceive timeout");
    return;
  }

  c = x10read(cm11, rbuf, 3);

  if(c <= 0) {
    if(c) {
      if(errno == EWOULDBLOCK)
	goto again;
      syslog(LOG_ERR, "read pollreceive: %m");
    }
    else
      syslog(LOG_ERR, "read pollreceive: EOF");
    return;
  }

retrysize:

  i = *rbuf & 0xff;

  if(i > 9 || i < 2) {
    if(i == POLL) {
      syslog(LOG_INFO, "pollreceive saw extra 0x5a (%d/%d)",
	     c, ecnt);
      if(c == 1)
      {
	if(ecnt++)
	  goto top;
	else
	  goto again;
      }

      c--;
      bcopy(rbuf + 1, rbuf, c);
      goto retrysize;
    }
    syslog(LOG_ERR, "pollreceive unreasonable upload size: %d", i);
    return;
  }

  i -= (j = c) - 1;		/* What the HELL is this? */

  while(i > 0) {
    FD_SET(cm11, &xprobe);
    tv.tv_sec = 2;

    c = select(cm11 + 1, &xprobe, FD_NULL, FD_NULL, &tv);

    if(c < 0) {
      syslog(LOG_ERR, "select2 pollreceive: %m");
      return;
    }
    if(c == 0) {
      syslog(LOG_ERR, "select2 pollreceive timeout");
      return;
    }
    c = x10read(cm11, rbuf + j, i);
    if(c <= 0) {
      if(c) {
	if(errno == EWOULDBLOCK)
	  continue;
	syslog(LOG_ERR, "read2 pollreceive: %m");
      }
      else
	syslog(LOG_ERR, "read2 pollreceive: EOF");
      return;
    }
    i -= c;
    j += c;
  }
  j = (*rbuf & 0xff) - 1;

  for(i = 0; i < j; i++) {
    c = rbuf[i + 2] & 0xf;

    if(rbuf[1] & (1 << i))
      c += 17;
    else
      c = fmap[c];

    if(c == X10_DIM || c == X10_BRIGHT)
      r = ((rbuf[i + 3] & 0xff) + 13) / 14;
    else
      r = 1;

    while(r--)
      processx(fmap[(rbuf[i + 2] >> 4) & 0xf] + 'A' - 1, /* House */
	       c,					 /* Func  */
	       0,					 /* Flag  */
	       rbuf[i + 3] & 0xff, 			 /* X1    */
	       rbuf[i + 4] & 0xff,                       /* X2    */
	       rbuf[i + 5] & 0xff);                      /* X3    */

    if(c == X10_DIM || c == X10_BRIGHT)
      i++;
    else if(c == X10_EXTENDED)
      i += 3;
  }
}

setcm11clock()
{
  register struct tm *tm;
  register int c;
  u_long t;
  u_char sum;
  struct cm11_setclock sc;
  static struct timeval tv;

  time(&t);
  tm = localtime(&t);

  sc.cmd     = TIME_DL;
  sc.seconds = tm->tm_sec;
  sc.minutes = tm->tm_min + 60 * (tm->tm_hour & 1);
  sc.hours   = tm->tm_hour >> 1;
  sc.yearday = tm->tm_yday;
  sc.daymask = 1 << tm->tm_wday;
  sc.house   = 0x60;


  if(tm->tm_yday & 0x100)
    sc.daymask |= 0x80;

  for(c = 1, sum = 0; c < sizeof(sc); c++)
    sum += ((u_char *)&sc)[c];

  x10write(cm11, (char *)&sc, sizeof(sc));

again:
  FD_SET(cm11, &xprobe);
  tv.tv_sec = 2;

  c = select(cm11 + 1, &xprobe, FD_NULL, FD_NULL, &tv);

  if(c < 0) {
    syslog(LOG_ERR, "select setcm11clock: %m");
    sleep(2);
    goto again;
  }
  if(c == 0) {
    syslog(LOG_ERR, "select setcm11clock timeout");
    return(-1);
  }

  c = x10read(cm11, rbuf, sizeof(rbuf));

  if(c <= 0) {
    if(c) {
      if(errno == EWOULDBLOCK)
	goto again;
      syslog(LOG_ERR, "read setcm11clock: %m");
    }
    else
      syslog(LOG_ERR, "read setcm11clock: EOF");
    sleep(1);
    return(-1);
  }
  if(c > 1)
    syslog(LOG_INFO, "setcm11clock %d byte response", c);
  c = rbuf[c - 1] & 0xff;
  if(c != sum) {
    syslog(LOG_ERR, "setcm11clock bad checksum %x expected %x",
	   c, sum);
    return(-1);
  }

  x10write(cm11, ACK, 1);

again2:

  FD_SET(cm11, &xprobe);
  tv.tv_sec = 2;

  c = select(cm11 + 1, &xprobe, FD_NULL, FD_NULL, &tv);

  if(c < 0) {
    syslog(LOG_ERR, "select2 setcm11clock: %m");
    sleep(2);
    goto again2;
  }
  if(c == 0) {
    syslog(LOG_ERR, "select2 setcm11clock timeout");
    return(-1);
  }
  c = x10read(cm11, rbuf, sizeof(rbuf));
  if(c <= 0) {
    if(c) {
      if(errno == EWOULDBLOCK)
	goto again2;
      syslog(LOG_ERR, "read2 setcm11clock: %m");
    }
    else
      syslog(LOG_ERR, "read2 setcm11clock: EOF");
    sleep(1);
    return(-1);
  }
  if(c > 1)
    syslog(LOG_INFO, "setcm11clock %d byte response2", c);
  c = *rbuf & 0xff;
  if(c != READY) {
    syslog(LOG_ERR, "setcm11clock bad ready %x", c);
    return(-1);
  }
  return(0);
}

char modstate[16][16];

processx(house, func, flag, x1, x2, x3)
{
  register int i, c;
  register char *s = modstate[house - 'A'];
  char buf[128];
  FILE *m;
  u_long t;

  if(func != X10_EXTENDED)
    x1 = x2 = 0;

  if(func <= 16) {
    for(i = 0; i < 16; i++)
      if(s[i] == 2)
	s[i] = 0;

    s[func - 1] = 1;
  }
  else {
    switch(func) {

      case X10_ALL_UNITS_OFF:
	bzero(s, 16);
	break;

      case X10_ON:
      case X10_OFF:
      case X10_DIM:
      case X10_BRIGHT:
      case X10_ALL_LIGHTS_ON:
	for(i = 0; i < 16; i++)
	  if(s[i] == 1)
	    s[i] = 2;
	break;
    }
  }


  for(i = c = 0; i < 16; i++)
    if(s[i])
      c |= (1 << i);

  if(func == X10_EXTENDED)
  {
      sprintf(buf, "%c %c %02x %02x %02x %s",
	      flag ? 'X' : 'R', house, x1, x2, x3, x10longnames[func]);
  }
  else
  {
    sprintf(buf, "%c %c %s", flag ? 'X' : 'R', house, x10longnames[func]);
  }

  syslog(LOG_DEBUG, buf);
  strcat(buf,"\r\n");
  c = strlen(buf);

  for(i = 0; i < nclient; i++) {
    if(clients[i].flags & C_QUIET)
      continue;

    if(write(clients[i].s, buf, c) != c)
      shutdown(clients[i].s, 2);
  }
}

/*
** A-P are already in use (X10 House codes)
**
** Q-Z are available
** Q   quiet mode
** R   repeat
** X   extended
** Z   Quit
** ?   Help
*/

parse(i, c)
register char *i;
int c;				/* c is the client in client[c] */
{
  register struct x10aliases *x;
  register int code, j;
  int repeat = 1, msglen;
  static struct timeval tv;
  u_char msg[5], sum;
  char string[80];

  while(*i)
  {

    if((strncasecmp(rbuf, "help", 4) == 0))
    {
      i += 4;
      help(c);
      continue;
    }
#ifdef NJC
    /*
    ** disconnect from this end
    ** This doesn cleanup properly, don't use it
    */

    if((strncasecmp(rbuf, "exit", 4) == 0))
    {
      i += 4;
      write(clients[c].s, "I \r\nI Bye!\r\n", 12);
      syslog(LOG_ERR, "Program exiting normally");
      shutdown(clients[c].s, 2);
      continue;
    }
#endif
    if(strncasecmp(rbuf, "status", 6) == 0)
    {
      i += 6;
      write(clients[c].s, "I Function currently not available\r\n", 36);
      continue;
    }

    if(strncasecmp(rbuf, "info", 4) == 0)
    {
      i += 4;
      write(clients[c].s, "I Function currently not available\r\n", 36);
      continue;
    }

    if(*i == 'Q') {		/* Quiet */
      i++;
      clients[c].flags ^= C_QUIET;
      continue;
    }

    if(*i == 'R') {		/* Repeat */
      i++;
      if(*i > '9' || *i < '0')
	goto syntax;
      repeat = 0;
      while(*i <= '9' && *i >= '0')
	repeat = 10 * repeat + *i++ - '0';
      if(!repeat)
	repeat = 1;
      if(repeat > 32)
	repeat = 32;
      continue;
    }

    if(*i == 'X') {		/* Extended Command */
      i++;
      if(strlen(i) < 6)
	goto syntax;
      sscanf(i, "%6x", &msglen);
      msg[2] = msglen >> 16;
      msg[3] = msglen >> 8;
      msg[4] = msglen;
      i += 6;
      continue;
    }

    if(*i < 'A' || *i > 'P') {	/* Not a command */
    syntax:
      write(clients[c].s, "E Syntax\r\n", 10);
      return;
    }

    msg[0] = ((repeat - 1) << 3) | 4;
    msg[1] = tmap[*i++ - 'A' + 1] << 4;

    if(*i <= '9' && *i >= '0') {
      code = 0;
      while(*i <= '9' && *i >= '0')
	code = 10 * code + *i++ - '0';
      if(!code || code > 32)
	goto syntax;
    }
    else
    {
      for(x = x10aliases; x->name; x++)
      {
	if(!memcmp(i, x->name, strlen(x->name)))
	{
	  break;
	}
      }

      if(!x->name)
      {
	goto syntax;
      }

      code = x->code;
      i += strlen(x->name);
    }

    if(code == X10_EXTENDED) {
      msg[0] |= 1;
      msglen = 5;
    }
    else
      msglen = 2;

    if(code <= 16)
      msg[1] |= tmap[code];
    else {
      msg[0] |= 2;
      msg[1] |= code - 17;
    }

    for(sum = j = 0; j < msglen; j++)
      sum += msg[j];
    
    /************************************************************************
    **
    ** Write the first 2 bytes wait for a checksum reply
    */

    x10write(cm11, msg, msglen);

  again:

    FD_SET(cm11, &xprobe);
    tv.tv_sec = 2;

    code = select(cm11 + 1, &xprobe, FD_NULL, FD_NULL, &tv);

    if(code < 0) {
      syslog(LOG_ERR, "select parse: %m");
      sleep(2);
      goto again;
    }

    if(code == 0) {
      syslog(LOG_ERR, "select parse timeout");
      write(clients[c].s, "E device timeout\r\n", 18);
      return;
    }

    /*
    ** Read in the checksum
    */

    code = x10read(cm11, rbuf, sizeof(rbuf));

    if(code <= 0)
    {
      if(code)
      {
	if(errno == EWOULDBLOCK)
	  goto again;
	syslog(LOG_ERR, "read parse: %m");
      }
      else
      {
	syslog(LOG_ERR, "read parse: EOF");
      }
      write(clients[c].s, "E bad device read\r\n", 19);
      sleep(1);
      return;
    }

    if(code > 1)
      syslog(LOG_INFO, "parse %d byte response", code);

    code = rbuf[code - 1] & 0xff;

    if(code != sum) {
      write(clients[c].s, "Bad checksum\r\n", 14);

      sprintf(string, "parse bad checksum %x expected %x", code, sum);
      syslog(LOG_INFO, string);

      if(code == FFail || code == POLL) {
	write(clients[c].s, "E transient fail\r\n", 18);
	return;
      }
      if(code == POLL)
	pollreceive();
      else
	sleep(1);

      x10write(cm11, msg, msglen);
      goto again;
    }

    /*
    ** Ack the checksum, if it's OK.
    */

    x10write(cm11, ACK, 1);

  again2:
    FD_SET(cm11, &xprobe);
    tv.tv_sec = 2 + 2 * repeat;

    code = select(cm11 + 1, &xprobe, FD_NULL, FD_NULL, &tv);

    if(code < 0) {
      syslog(LOG_ERR, "select2 parse: %m");
      sleep(2);
      goto again2;
    }
    if(code == 0) {
      syslog(LOG_ERR, "select2 parse timeout");
      write(clients[c].s, "E device timeout\r\n", 18);
      return;
    }

    code = x10read(cm11, rbuf, sizeof(rbuf));

    if(code <= 0) {
      if(code) {
	if(errno == EWOULDBLOCK)
	  goto again2;
	syslog(LOG_ERR, "read2 parse: %m");
      }
      else
	syslog(LOG_ERR, "read2 parse: EOF");
      write(clients[c].s, "E bad device read\r\n", 19);
      sleep(1);
      return;
    }
    if(code > 1)
      syslog(LOG_INFO, "parse %d byte response2", code);
    code = *rbuf & 0xff;

    if(code != READY) {
      syslog(LOG_INFO, "parse bad ready %x", code);
      if(code == FFail || code == PFail) {
	write(clients[c].s, "E transient fail\r\n", 18);
	return;
      }
      if(code == POLL)
	pollreceive();
      else
	sleep(1);
      x10write(cm11, msg, msglen);
      goto again;
    }
    code = msg[1] & 0xf;
    if(msg[0] & 2)
      code += 17;
    else
      code = fmap[code];
    j = ((msg[0] >> 3) & 0x1f) + 1;
    if(j > 1 && !((j-2)%3))
      j--;

    while(j--)
      processx(fmap[(msg[1] >> 4) & 0xf] + 'A' - 1,
	       code,
	       1,
	       msg[2],
	       msg[3],
	       msg[4]);
    /*
    ** I think we're done
    ***********************************************************/
  }
  write(clients[c].s, "A\r\n", 3);
}

/*
** x10write & x10read are nothing more than intercepts
** for read and write. It allows me to keep track of the bytes
** being sent and received by the CM11 device.
*/

ssize_t x10write(int device, char *buffer, size_t count)
{
  static char string[80];
  int  i;

  strcpy(string, "Write:");
  for(i = 0; i < count; i++)
    sprintf(string, "%s 0x%02x", string, MASK(buffer[i]));

  syslog(LOG_DEBUG, string);

  return(write(device, buffer, count));
}

ssize_t x10read(int device, char *buffer, size_t count)
{
  static char string[80];
  int  i, x;

  strcpy(string, "Read: ");

  if(!((x = read(device, buffer, count)) < 0))
  {
    for(i = 0; i < x; i++)
      sprintf(string, "%s 0x%02x", string, MASK(buffer[i]));
  }
  else
  {
    sprintf(string, "%s(%d)", string, x);
  }

  syslog(LOG_DEBUG, string);
    
  return(x);
}

void help(int c)
{
  write(clients[c].s, "\r\nH Help message\r\n\n", 18);
#ifdef NJC
  write(clients[c].s, "H exit   - This help screen\r\n", 29);
#endif
  write(clients[c].s, "H info   - This help screen\r\n", 29);
  write(clients[c].s, "H status - This help screen\r\n", 29);
  write(clients[c].s, "H  ?     - This help screen\r\n", 29);
  write(clients[c].s, "H  Q     -\r\n", 12);
  write(clients[c].s, "H  X     -\r\n", 12);
  write(clients[c].s, "H  Z     - Quit\r\nH\r\n", 20);
  write(clients[c].s, "H Example command A1AON turns on unit A1\r\n", 42);
  write(clients[c].s, "\r\n", 2);
}

SignalHandler fini()
{
  syslog(LOG_DEBUG,"Fini\n");
  exit(0);
}
