#!/bin/sh
#

usage()
{
  echo "Usage: on|off|dim|bright|setback|setforward device"
  echo "Ex:"
  echo "    on tv"
  echo "or  on p15"
  echo
}

# LIST is a file with a list of aliases and device codes separated by a TAB (1)
# Field 1 is an alias, Field 2 is the Device code

if [ "X${LIST}" = "X" ] ; then
  LIST="${HOME}/.x10alias"
fi

# HOST is the hostname of the machine running the X10D daemon, HOST can also
# be any valid IP address

if [ "X${HOST}" = "X" ] ; then
  HOST="diane"
fi

# PORT is the port name of the X10D (name found in /etc/services), PORT can
# also be a valid PORT number

if [ "X${PORT}" = "X" ] ; then
  PORT="x10d"
fi

# This takes care of anything calling this program from x10talk.sh
# instead of by one of its links

NAME=`basename ${0}`

if [ "${NAME}" = "x10talk.sh" ] ; then
  NAME=`basename ${1}`			# $1 is the X10 command
  shift					# a shift won't change $0
fi

# catch too few arg error

if [ $# = 0 ] ; then
  usage
  exit 1
fi

if [ -e "$LIST" ]; then
  # cat the list,
  # save everything up to the #,
  # check for alias, save to UNIT

  UNIT=`cat ${LIST}|cut -f 1 -d \#|grep ^${1}|cut -f2`
fi

# The House code is the first letter of the UNIT

HOUSE=`echo ${UNIT:-${1}}|cut -b1`

# the :-${1} handles an arg1 of a15 (a1 - p16, 256 devices)
# which won't be found in the .x10alias file

UNIT=`echo ${UNIT:-${1}}|cut -f1`

UC="${UNIT}${HOUSE}"

#echo Fiddling with ${1} \(${UNIT}\)

case "${NAME}" in

"on"|"setback")
        echo Turning ${1} \(${UNIT}\) on.
        hose ${HOST} ${PORT} -in -out \
                sh -c " ( echo "${UC}ON" ;  sockdown ) & "
        ;;
  
"off"|"setforward")
        echo Turning ${1} \(${UNIT}\) off.
        hose ${HOST} ${PORT} -in -out \
                sh -c " ( echo "${UC}OFF" ;  sockdown ) & "
        ;;

"bright")
        echo Brightening ${1} \(${UNIT}\).
        hose ${HOST} ${PORT} -in -out \
                sh -c " ( echo "${UC}BRIGHT" ;  sockdown ) & "
        ;;

"dim")
        echo Dimming ${1} \(${UNIT}\).
        hose ${HOST} ${PORT} -in -out \
                sh -c " ( echo "${UC}DIM" ;  sockdown ) & "
        ;;

#"status")
#	echo Getting the status for ${1} \(${UNIT}\).
#	hose ${HOST} ${PORT} -in -out \
#		sh -c " ( echo "${UC}DIM" ;  sockdown ) & "
#	;;

*)
	usage
        echo "Usage:"
        echo "on|off|dim|bright|setback|setforward device"
        echo "Ex: on tv"
        echo "or  on p15"
        ;;
esac

exit 0
